
/**
 * @overview Defines the permissions provider class.
 *
 * @module stout-auth/provider/PermissionsProvider
 */

(function() {
  var Provider, err, has, intersection, isArray, isEqual, set;

  err = require('stout-core/err');

  has = require('lodash/has');

  intersection = require('lodash/intersection');

  isArray = require('lodash/isArray');

  isEqual = require('lodash/isEqual');

  Provider = require('stout-core/provider/Provider');

  set = require('lodash/set');


  /**
   *
   *
   *
   */

  module.exports = Provider.extend('PermissionsProvider', {
    constructor: function() {
      Provider.prototype.constructor.call(this);
      this.__permissions = [];
      this.__actions = {};
    },

    /**
     * Boot this provider before other provider so it's available to define
     * permissions at boot time.
     *
     * @member {number} $bootOrder
     * @memberof stout-core/provider/Provider
     */
    $bootOrder: 1900,

    /**
     * Defines a new permissions. Grouped permissions should be defined using the
     * dot-notation.
     *
     * @param {string} permission - The permission name, i.e. `events.publish`.
     *
     * @throws {stout-core/err/Err} Thrown if the permission is already defined.
     *
     * @method define
     * @memberof stout-auth/provider/PermissionsProvider
     */
    define: function(permission) {
      if (this.defined(permission)) {
        throw new err.Err("Permission " + permission + " already defined.");
      }
      return this.__permissions.push(permission);
    },

    /**
     * Defines a plain-english action which may require a set of permissions. When
     * testing if an action is allowed, the defined function `fn` is called with
     * an array of permissions. The function `fn` should return `true` or `false`
     * if the action is allowed given the set of permissions.
     *
     * Alternatively, an array of permissions may be passed as `fn`. When testing
     * if an action should be allowed, if all specified permissions (`fn`) are
     * present in the list of testing permissions, the action is allowed.
     *
     * @method defineAction
     * @memberof stout-auth/provider/PermissionsProvider#
     */
    defineAction: function(action, fn) {
      if (isArray(fn)) {
        fn = (function(reqPerms) {
          reqPerms.sort();
          return function(permissions) {
            return isEqual(reqPerms, intersection(reqPerms, permissions).sort());
          };
        })(fn);
      }
      this.__actions[action] = fn;
      return fn;
    },

    /**
     * Checks if a give permission is defined.
     *
     * @param {string} permission - The permission name, i.e. `events.publish`.
     *
     * @returns {boolean} `true` if the permission is defined, otherwise `false`.
     *
     * @method defined
     * @memberof stout-auth/provider/PermissionsProvider
     */
    defined: function(permission) {
      return this.__permissions.indexOf(permission) !== -1;
    },

    /**
     * Returns the testing-function for the specified action. If that action does
     * not exist, an error is thrown.
     *
     * @param {string} action - The plain-english action.
     *
     * @method getAction
     * @memberof stout-auth/provider/PermissionsProvider#
     */
    getAction: function(action) {
      var fn;
      fn = this.__actions[action];
      if (!fn) {
        throw new err.Err("The action \"" + action + "\" is not defined.");
      }
      return fn;
    },

    /**
     * Tests if the specified action is allowed given the passed set of
     * permissions.
     *
     * @method testAction
     * @memberof stout-auth/provider/PermissionsProvider
     */
    testAction: function(action, permissions) {
      return this.getAction(action)(permissions);
    }
  });

}).call(this);

//# sourceMappingURL=PermissionsProvider.js.map
