
/**
 * @overview Defines the default server-side authentication services class which
 * accepts and processes authentication requests from a client using the
 * `stout-auth/service/ClientAuthService` service.
 *
 * @module stout-auth/service/ServerAuthService
 */

(function() {
  var AuthService, Password, PasswordReset, Promise, Scrypt, crypto, err;

  AuthService = require('./AuthService');

  crypto = require('crypto');

  err = require('../err');

  Password = require('../model/Password');

  PasswordReset = require('../model/Password');

  Promise = require('stout-core/promise/Promise');

  Scrypt = require('../password/Scrypt');


  /**
   * The ServerAuthService class provides server-side authentication services for
   * web applications. It's designed to seamlessly handle requests originating
   * from clients running the corresponding client-side service,
   * `stout-auth/service/ClientAuthService`.
   *
   * @exports stout-auth/service/ServerAuthService
   * @extends stout-auth/service/AuthProvider
   * @constructor
   */

  module.exports = AuthService.extend('ServerAuthService', {
    constructor: function() {
      AuthService.prototype.constructor.apply(this, arguments);
      this.hasher = new Scrypt();
      this.handlers = {
        'signin': this.signin,
        'signout': this.signout,
        'generate-verification-token': this.generateVerificationToken,
        'check-verification-token': this.checkVerificationToken,
        'register': this.register,
        'request-password-reset': this.requestPasswordReset,
        'reset-password': this.resetPassword,
        'change-password': this.changePassword
      };
    },

    /**
     * Checks if the `password` matches the user's password.
     *
     * @param {stout-auth/model/User} user - The user model.
     *
     * @param {string} password - The password to check.
     *
     * @returns {stout-core/promise/Promise} Promise resolved to `true` or `false`,
     * indicating if the password matches, or rejected if the user does not have
     * an associated password
     *
     * @method _checkPassword
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _checkPassword: function(user, rawPassword) {
      return user.password.get().then((function(_this) {
        return function(password) {
          if (!password) {
            throw new err.AuthErr('User does not have a password.');
          } else {
            return _this._getHasher(password).verify(password.verifier, rawPassword);
          }
        };
      })(this));
    },

    /**
     * Returns the appropriate hasher for the passed password object. Currently,
     * this method simply returns the only hasher, but it could be extended in the
     * future to return one of multiple hashers if the database contains passwords
     * hashed over time with separate passwords.
     *
     * @method _getHasher
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _getHasher: function(password) {
      return this.hasher;
    },

    /**
     * Returns the reduced precision clock for generating and checking
     * verification tokens.
     *
     * @returns {number} Reduced precision clock.
     *
     * @method _getVerificationClock
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _getVerificationClock: function() {
      var config, period;
      config = this.providers.get('config');
      period = config.get('auth.verification.period') * 60 * 1000 / 2;
      return parseInt(Date.now() / period);
    },

    /**
     * Makes a new password reset model for the passed user.
     *
     * @param {stout-auth/model/User} user - The user to generate the password
     * reset for.
     *
     * @returns {stout-auth/model/PasswordReset} The password reset model.
     *
     * @method _makePasswordReset
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _makePasswordReset: function(user) {
      var config, expires, hmac, secret;
      config = this.providers.get('config');
      secret = config.get('auth.secret');
      expires = config.get('auth.reset.expires');
      hmac = crypto.createHmac('sha256', secret);
      hmac.update(user.key);
      return new PasswordReset({
        token: hmac.digest('hex'),
        expires: Date.now() + expires * 60 * 1000
      });
    },

    /**
     * Generates a verification token for the passed user and session.
     *
     * @param {stout-auth/model/User} user - The user model.
     *
     * @param {} session - The user's current session.
     *
     * @returns {string} The verification token.
     *
     * @method _makeVerificationToken
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _makeVerificationToken: function(user, session, clock) {
      var config, hmac, secret;
      config = this.providers.get('config');
      secret = config.get('auth.secret');
      hmac = crypto.createHmac('sha256', secret);
      hmac.update(user.key + clock + session.key);
      return hmac.digest('hex');
    },

    /**
     * Generates a new password verifier and returns a Password model.
     *
     * @param {string} password - The password to hash.
     *
     * @returns {stout-core/promise/Promise} Promise resolved with Password model.
     *
     * @method _makePasswordVerifier
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _makePasswordVerifier: function(password) {
      return this.hasher.generate(password).then(function(verifier) {
        var pass;
        pass = new Password({
          algorithm: 'scrypt',
          version: 1,
          verifier: verifier
        });
        return pass.save().then(function() {
          return pass;
        });
      });
    },

    /**
     * Sends the password reset request email to the user.
     *
     * @param {stout-auth/model/User} user - The user to send the password
     * reset request email to.
     *
     * @param {stout-auth/model/passwordReset} passwordReset - The password reset
     * model.
     *
     * @method _sendPasswordResetEmail
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _sendPasswordResetEmail: function(user, passwordReset) {
      return this.services.request('email:send', {
        template: 'emails.password-reset',
        date: {
          user: user,
          passwordReset: passwordReset
        },
        to: user.email
      });
    },

    /**
     * Creates a password verifier and attaches it to the passed user model.
     *
     * @param {stout-auth/model/User} user - The user model.
     *
     * @param {string} password - The password.
     *
     * @method _setUserPassword
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _setUserPassword: function(user, password) {
      return this._makePasswordVerifier(password).then(function(verifier) {
        return user.password.set(verifier).then(function() {
          return user.save();
        })["catch"](function(error) {
          verifier["delete"]();
          throw error;
        });
      });
    },

    /**
     * Starts a session for the passed user.
     *
     * @method _startSession
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _startSession: function(user) {
      return Promise.fulfilled();
    },

    /**
     * Ends a session for the passed user.
     *
     * @method _endSession
     * @memberof stout-auth/server/AuthProvider#
     * @protected
     */
    _endSession: function(user) {
      return Promise.fulfilled();
    },

    /**
     * Changes a user's password.
     *
     * @param {function} fulfill - Function to call after a successful password
     * change.
     *
     * @param {function} reject - Function to call with a failed password change.
     *
     * @param {stout-auth/model/User} user - The user object.
     *
     * @param {string} currentPassword - The user's current password.
     *
     * @param {string} password - The new password.
     *
     * @method changePassword
     * @memberof stout-auth/server/AuthProvider#
     */
    changePassword: function(fulfill, reject, user, currentPassword, newPassword) {
      var username;
      username = user.getUsername();
      this.log.debug('Changing password for user %s.', username, user);
      this.fire('change-password:request', user);
      return this._checkPassword(user, currentPassword).then(function(matches) {
        var error;
        if (matches) {
          return this._setUserPassword(user, newPassword).then(function() {
            this.log.info('Successful changed password for user %s.', username, user);
            this.fire('change-password:success', user);
            return fulfill();
          });
        } else {
          this.log.notice('Failed to change password for user %s. Current password does not match.', username, user);
          error = new err.PasswordErr('Invalid password.');
          this.fire('change-password:error', {
            user: user,
            error: error
          });
          return reject(error);
        }
      })["catch"](function(error) {
        this.log.warning('Error occurred attempting to change password for user %s.', username, {
          user: user,
          error: error
        });
        this.fire('change-password:error', {
          user: user,
          error: error
        });
        return reject(error);
      });
    },

    /**
     * Checks a verification token. If the token is valid, the request is
     * fulfilled with an updated token.
     *
     * @param {function} fulfill - Function to call with a successful verification.
     *
     * @param {function} reject - Function to call with a failed verification.
     *
     * @param {stout-auth/model/User} user - The user object.
     *
     * @param {string} token - The token to check.
     *
     * @method checkVerificationToken
     * @memberof stout-auth/server/AuthProvider#
     */
    checkVerificationToken: function(fulfill, reject, user, token) {
      return user.session.get().then(function(session) {
        var clock, error, vToken, vTokenPrevious;
        if (!session) {
          this.log.warning('Cannot check token for user %s. No valid session exists.', username, user);
          error = new err.VerificationErr('No valid session exists.');
          return this.fire('verify:error', {
            user: user,
            error: error
          });
        } else {
          clock = this._getVerificationClock();
          vToken = this._makeVerificationToken(user, session, clock);
          vTokenPrevious = this._makeVerificationToken(user, session, clock - 1);
          if (vToken === token || vTokenPrevious === token) {
            this.log.info('Validated verification token for user %s.', user);
            this.fire('verify:success', {
              user: user,
              token: token
            });
            return fulfill(vToken);
          } else {
            this.log.info('Invalid verification token for user %s.', user);
            this.fire('verify:success', {
              user: user,
              token: token
            });
            return fulfill();
          }
        }
      })["catch"](function(error) {
        this.log.warning('Error occurred attempting to check verification token for %s.', username, {
          user: user,
          error: error
        });
        this.fire('verify:error', {
          user: user,
          error: error
        });
        return reject(error);
      });
    },

    /**
     * Verifies the passed credentials and returns a token and timestamp for later
     * use to ensure the user did in-fact verify their credentials within a given
     * amount of time.
     *
     * Here's how this works:
     * 1. Alice performs some operation (client side) which requires that she
     *    verify her password (i.e. change critical settings, delete account,
     *    etc.).
     * 2. Alice is presented with a "verify password" modal or similar UI element
     *    where she enters her password.
     * 3. The client-side system sends the password to the server.
     * 4. The server checks if the password matches Alice's stored password.
     * 5. If the passwords match, a token is generated as follows:
     *    a. Let `id` be Alice's user ID (UUID), `t` be the current timestamp,
     *       and `secret` be a secret key stored on the server.
     *    b. Let `k = HMAC(secret, id + t)`
     *    c. Generate a secure token as `token = k + t`,
     *       where `+` is the concatenation operator.
     * 6. Send the token to Alice.
     * 7. For each "dangerous" operation, Alice includes `token` with the request.
     * 8. The server can verify the `token` by splitting `k` and `t` and
     * validating `k` as in step 5.b above, and verify that time `T` hasn't
     * elapsed since Alice verified her password.
     *
     * Alice can then include the token with future dangerous requests to validate
     * that she has verified her password, and that it was verified at a specific
     * time. The time can be used to ensure that when performing dangerous
     * operations, the user has verified their identity within the past X minutes.
     *
     * @param {object} [request] - Credentials verification request.
     *
     * @param {function} fulfill - Function to call with a successful verify.
     *
     * @param {function} reject - Function to call with a failed verify.
     *
     * @method generateVerificationToken
     * @memberof stout-auth/server/AuthProvider#
     */
    generateVerificationToken: function(fulfill, reject, user, password) {
      return this._checkPassword(user, password).then(function(matches) {
        var error;
        if (matches) {
          return user.session.get().then(function(session) {
            var clock, error, token;
            if (!session) {
              this.log.warning('Cannot generate token for user %s. No valid session exists.', username, user);
              error = new err.VerificationErr('No valid session exists.');
              return this.fire('verify:token:error', {
                user: user,
                error: error
              });
            } else {
              clock = _getVerificationClock();
              token = this._makeVerificationToken(user, session, clock);
              this.log.info('Generated verification token for %s.', user);
              this.fire('verify:token:success', {
                user: user,
                token: token
              });
              return fulfill(token);
            }
          });
        } else {
          this.log.notice('Failed verification token generation attempt. Incorrect password for user %s.', username, user);
          error = new err.PasswordErr('Invalid password.');
          this.fire('verify:token:error', {
            user: user,
            error: error
          });
          return reject(error);
        }
      })["catch"](function(error) {
        this.log.warning('Error occurred attempting to generate verification token for %s.', username, {
          user: user,
          error: error
        });
        this.fire('verify:token:error', {
          user: user,
          error: error
        });
        return reject(error);
      });
    },

    /**
     * Registers a new user by hashing the password, then saving the user and
     * linking the password verifier.
     *
     * @param {object} [request] - Registration request.
     *
     * @param {function} fulfill - Function to call with a successful registration.
     *
     * @param {function} reject - Function to call with a failed registration.
     *
     * @method register
     * @memberof stout-auth/server/AuthProvider#
     */
    register: function(fulfill, reject, user, password) {
      var username;
      username = user.getUsername();
      this.log.debug('Registering user %s.', username, user);
      this.fire('register:request');
      return this._makePasswordVerifier(password).then((function(_this) {
        return function(verifier) {
          return user.password.link(verifier).then(function() {
            return user.save().then(function() {
              _this.log.info('Successfully registered user %s.', username, user);
              _this.fire('register:success', user);
              return fulfill();
            });
          })["catch"](function(error) {
            _this.log.warning('Error occurred registering %s.', username, {
              user: user,
              error: error
            });
            _this.fire('register:error', {
              user: user,
              error: error
            });
            verifier["delete"]();
            user["delete"]();
            return reject(error);
          });
        };
      })(this))["catch"]((function(_this) {
        return function(error) {
          _this.log.warning('Error generating password verifier for %s.', username, {
            user: user,
            error: error
          });
          _this.fire('register:error', {
            user: user,
            error: error
          });
          return reject(error);
        };
      })(this));
    },

    /**
     * Deletes any pending password resets for the passed user, generates a new
     * password reset request, and emails the user the password reset email.
     *
     * @param {function} fulfill - Function to call with a successful password
     * reset request.
     *
     * @param {function} reject - Function to call with a failed password reset
     * request.
     *
     * @param {stout-auth/model/User} user - The user object.
     *
     * @method requestPasswordReset
     * @memberof stout-auth/server/AuthProvider#
     */
    requestPasswordReset: function(fulfill, reject, user) {
      var username;
      username = user.getUsername();
      return user.passwordReset.get().then(function(previousReset) {
        var passwordReset;
        if (previousReset) {
          previousReset["delete"]()["catch"](function(error) {
            return this.log.error('Error deleting previous password reset for user %s.', username, {
              user: user,
              error: error
            });
          });
        }
        passwordReset = this._makePasswordReset(user);
        return user.passwordReset.set(passwordReset).then(function() {
          return this._sendPasswordResetEmail(user, passwordReset).then(function() {
            this.log.info('Sent password reset email to user %s.', username, user);
            return fulfill();
          })["catch"](function(error) {
            this.log.warning('Error sending password reset email to user %s.', username, {
              user: user,
              error: error
            });
            return reject(error);
          });
        });
      })["catch"](function(error) {
        this.log.warning('Error generating password reset for user %s.', username, {
          user: user,
          error: error
        });
        return reject(error);
      });
    },

    /**
     * Resets the user's password.
     *
     * @param {function} fulfill - Function to call after a successful password
     * reset.
     *
     * @param {function} reject - Function to call with a failed password reset.
     *
     * @param {stout-auth/model/User} user - The user object.
     *
     * @param {string} token - Password reset request token.
     *
     * @param {string} password - The new password.
     *
     * @method resetPassword
     * @memberof stout-auth/server/AuthProvider#
     */
    resetPassword: function(fulfill, reject, user, token, password) {
      var username;
      username = user.getUsername();
      this.log.debug('Resetting password for user %s.', username, user);
      return user.passwordReset.get().then(function(passwordReset) {
        var error;
        if (!passwordReset || passwordReset.expired() || passwordReset.token(!token)) {
          this.log.notice('Invalid password token or token has expired for user %s', username, user);
          error = new err.PasswordResetErr('Invalid or expired password reset token.');
          return reject(error);
        } else {
          return this._setUserPassword(user, password).then(function() {
            this.log.info('Successfully reset password for user %s.', username, user);
            this.fire('change-password:success', user);
            return fulfill();
          });
        }
      })["catch"](function(error) {
        this.log.warning('Error reseting password for user %s.', username, {
          user: user,
          error: error
        });
        return reject(error);
      });
    },

    /**
     * Signs-in a user, verifying the passed credentials.
     *
     * @param {object} [request] - Sign-in request.
     *
     * @param {function} fulfill - Function to call with a successful sign-in
     * request.
     *
     * @param {function} reject - Function to call with a failed sign-in
     * request.
     *
     * @method signin
     * @memberof stout-auth/server/AuthProvider#
     */
    signin: function(fulfill, reject, user, password) {
      var username;
      username = user.getUsername();
      this.log.debug('Signing-in user %s.', username, user);
      this.fire('signin:request', user);
      return this._checkPassword(user, password).then((function(_this) {
        return function(matches) {
          var error;
          if (matches) {
            _this.log.info('Successful sign in for %s.', username, user);
            _this.fire('signin:success', user);
            return _this._startSession(user).then(function() {
              return fulfill();
            });
          } else {
            _this.log.notice('Failed sign in attempt. Incorrect password for user %s.', username, user);
            error = new err.PasswordErr('Invalid password.');
            _this.fire('signin:error', {
              user: user,
              error: error
            });
            return reject(error);
          }
        };
      })(this))["catch"]((function(_this) {
        return function(error) {
          _this.log.warning('Error occurred attempting to sign in %s.', username, {
            user: user,
            error: error
          });
          _this.fire('signout:error', {
            user: user,
            error: error
          });
          return reject(error);
        };
      })(this));
    },

    /**
     * Signs-out a user.
     *
     * @param {object} [request] - Sign-out request.
     *
     * @param {function} fulfill - Function to call with a successful sign-out.
     *
     * @param {function} reject - Function to call with a failed sign-out.
     *
     * @method signout
     * @memberof stout-auth/server/AuthProvider#
     */
    signout: function(fulfill, reject, user) {
      var username;
      username = user.getUsername();
      this.log.debug('Signing-out user %s.', username, user);
      this.fire('signout:request');
      return this._endSession(user).then(function() {
        this.log.info('Successful sign out for %s.', username, user);
        this.fire('signout:success');
        return fulfill();
      })["catch"](function(error) {
        this.log.warning('Error occurred attempting to sign out %s.', username, {
          user: user,
          error: error
        });
        this.fire('signout:error', {
          user: user,
          error: error
        });
        return reject(error);
      });
    }
  });

}).call(this);

//# sourceMappingURL=ServerAuthService.js.map
