###*
# @overview Defines the base User class.
#
# @module stout-auth/model/User
###
isArray = require 'lodash/isArray'
Model   = require 'stout-orm/model/Model'
remove  = require 'lodash/remove'
union   = require 'lodash/union'
uniq    = require 'lodash/uniq'



###*
# The User class defines the minimal user model for compatibility with the
# stout-auth package.
#
# @exports stout-auth/model/User
# @extends stout-orm/model/Model
# @constructor
###
module.exports = Model.define 'User',

  ###*
  # Boots the User model and obtains reference to the PermissionsProvider.
  #
  # @param {stout-core/app/App} app - The application.
  #
  # @member {function} $boot
  # @memberof stout-core/provider/Provider#
  ###
  $boot: (app) ->
    Model.$boot.call @, arguments...
    @::_permissionsProvider = app.providers.get 'PermissionsProvider'


  properties:

    ###*
    # This user's email address.
    #
    # @member {string} email
    # @memberof stout-auth/model/User#
    ###
    email:
      type: 'string'
      required: true

    ###*
    # Set of permissions this user is granted.
    #
    # @member {string[]} permissions
    # @memberof stout-auth/model/User#
    ###
    permissions:
      default: []
      type: 'array'


  ###*
  # The provider name used to retrieve and store user model. The actual
  # provider class used is determined at app boot time.
  #
  # @member {mixed} provider
  # @memberof stout-auth/model/User#
  ###
  provider: 'UserProvider'


  ###*
  # Each user has an associated Password model which stores the password
  # verifier (it's hashed value), and verifier generation metadata. The
  # actual Password class used is determined at app boot time.
  #
  # @member {mixed} hasOne
  # @memberof stout-auth/model/User#
  ###
  hasOne: 'Password'


  ###*
  # Associated models in a has-many relationship. User models have the following
  # has-many type relationships.
  # - **Role** – Each user may be assigned zero or more roles which grant them
  #   a set of permissions defined in the role.
  #
  # @member {mixed} hasMany
  # @memberof stout-auth/model/User#
  ###
  hasMany: ['Role', 'Session']


  ###*
  # Returns the "username" property for this model, the property used for
  # logging in.
  #
  # @method getUsername
  # @memberof stout-auth/model/User#
  ###
  getUsername: -> @email


  ###*
  # Tests if this user is allowed to perform the specified action or has the
  # passed permission.
  #
  # @param {string} action - The action or permission to test.
  #
  # @method can
  # @memberof stout-auth/model/User#
  ###
  can: (action) ->
    @getAllPermissions().then (perms) =>
      @_permissionsProvider.testAction(action, perms)


  ###*
  # Returns the union of all individually-assigned and role-assigned
  # permissions.
  #
  # @returns {stout-core/promise/Promise} Promise resolved with all permissions.
  ###
  getAllPermissions: ->
    @roles.get().then (roles) =>
      perms = []
      for role in roles
        perms = perms.concat(role.permissions)
      uniq(union(@permissions, perms))


  ###*
  # Grants the passed permissions to this user. Passed permissions may be a
  # string or array of strings.
  #
  # @param {...string|...Array.<string>} permissions - The permissions to add.
  #
  # @method grant
  # @memberof stout-auth/model/User#
  ###
  grant: (permissions...) ->
    for perm in permissions
      if isArray(perm)
        @grant.apply @, perm
      else
        @permissions.push perm


  ###*
  # Checks if this user has the specified permission.
  #
  # @param {string} permission - The permission to check for.
  #
  # @returns {stout-core/promise/Promise} Promise resolved to `true` or `false`
  # indicating if the user has the specified permission.
  #
  # @method granted
  # @memberof stout-auth/model/User#
  ###
  granted: (permission) ->
    @getAllPermissions().then (perms) ->
      permission in perms


  ###*
  # Revokes the passed permissions from this user. Passed permissions may be a
  # string or array of strings. This will only revoke specific permissions
  # granted to this user. Permissions granted through roles assigned to this
  # user will not be affected.
  #
  # @param {...string|...Array.<string>} permissions - The permissions to
  # revoke.
  #
  # @method revoke
  # @memberof stout-auth/model/User#
  ###
  revoke: (permissions...) ->
    for perm in permissions
      if isArray(perm)
        @revoke.apply @, perm
      else
        remove @permissions, (existingPermission) ->
          existingPermission is perm
