###*
# @overview A password verifier generating and verifying class which uses the
# scrypt cryptographic hashing utility.
#
# @module stout-auth/password/Scrypt
###
scrypt  = require 'scrypt'
Promise = require 'stout-core/promise/Promise'



###*
# This class generates and verifies password verifiers using the scrypt
# cryptographic hash utility.
#
# @param {number} [maxtime=0.1] - A decimal (double) representing the maximum
# amount of time in seconds scrypt will spend when computing the derived key.
#
# @param {number} [maxmem=0] - An integer, specifying the maximum number of
# bytes of RAM used when computing the derived encryption key. If not present,
# will default to 0.
#
# @param {number} [maxmemfrac=0.5] - A double value between 0.0 and 1.0,
# representing the fraction (normalized percentage value) of the available
# RAM used when computing the derived key. If not present, will default to 0.5.
#
# @exports stout-auth/password/Scrypt
# @constructor
###
module.exports = class Scrypt

  constructor: (maxtime = 0.1, maxmem, maxmemfrac) ->
    @__params = scrypt.paramsSync(maxtime, maxmem, maxmemfrac)


  ###*
  # Generates a password verifier string, also known as a password hash.
  #
  # @param {string} password - The password the verify.
  #
  # @returns {stout-core/promise/Promise} Promise resolved with the generated
  # verifier, or rejected if an error occurs.
  #
  # @method generate
  # @memberof stout-auth/password/Scrypt
  ###
  generate: (password) ->
    new Promise (fulfill, reject) =>
      scrypt.kdf new Buffer(password), @__params, (err, result) ->
        if err
          reject(err)
        else
          fulfill(result.toString('base64'))


  ###*
  # Checks if the password matches the verifier (hash).
  #
  # @param {string} verifier - The password verifier (a.k.a it's hash).
  #
  # @param {string} password - The password the verify.
  #
  # @returns {stout-core/promise/Promise} Promise resolved with `true` or
  # `false` indicating if the password matches, and rejected if an error occurs.
  #
  # @method verify
  # @memberof stout-auth/password/Scrypt
  ###
  verify: (verifier, password) ->
    new Promise (fulfill, reject) ->
      decodedVerifier = new Buffer(verifier, 'base64')
      scrypt.verifyKdf decodedVerifier, password, (err, result) ->
        if err
          reject(err)
        else
          fulfill(result)
