###*
# @overview Defines the permissions provider class.
#
# @module stout-auth/provider/PermissionsProvider
###
err      = require 'stout-core/err'
has      = require 'lodash/has'
intersection = require 'lodash/intersection'
isArray  = require 'lodash/isArray'
isEqual  = require 'lodash/isEqual'
Provider = require 'stout-core/provider/Provider'
set      = require 'lodash/set'



###*
#
#
#
###
module.exports = Provider.extend 'PermissionsProvider',

  constructor: ->
    Provider.prototype.constructor.call @
    @__permissions = []
    @__actions = {}
    return


  ###*
  # Boot this provider before other provider so it's available to define
  # permissions at boot time.
  #
  # @member {number} $bootOrder
  # @memberof stout-core/provider/Provider
  ###
  $bootOrder: 1900


  ###*
  # Defines a new permissions. Grouped permissions should be defined using the
  # dot-notation.
  #
  # @param {string} permission - The permission name, i.e. `events.publish`.
  #
  # @throws {stout-core/err/Err} Thrown if the permission is already defined.
  #
  # @method define
  # @memberof stout-auth/provider/PermissionsProvider
  ###
  define: (permission) ->
    if @defined(permission)
      throw new err.Err "Permission #{permission} already defined."
    @__permissions.push permission


  ###*
  # Defines a plain-english action which may require a set of permissions. When
  # testing if an action is allowed, the defined function `fn` is called with
  # an array of permissions. The function `fn` should return `true` or `false`
  # if the action is allowed given the set of permissions.
  #
  # Alternatively, an array of permissions may be passed as `fn`. When testing
  # if an action should be allowed, if all specified permissions (`fn`) are
  # present in the list of testing permissions, the action is allowed.
  #
  # @method defineAction
  # @memberof stout-auth/provider/PermissionsProvider#
  ###
  defineAction: (action, fn) ->
    if isArray(fn)
      fn = ((reqPerms) ->
        reqPerms.sort()
        (permissions) ->
          isEqual(reqPerms, intersection(reqPerms, permissions).sort())
      )(fn)
    @__actions[action] = fn
    fn


  ###*
  # Checks if a give permission is defined.
  #
  # @param {string} permission - The permission name, i.e. `events.publish`.
  #
  # @returns {boolean} `true` if the permission is defined, otherwise `false`.
  #
  # @method defined
  # @memberof stout-auth/provider/PermissionsProvider
  ###
  defined: (permission) ->
    @__permissions.indexOf(permission) isnt -1


  ###*
  # Returns the testing-function for the specified action. If that action does
  # not exist, an error is thrown.
  #
  # @param {string} action - The plain-english action.
  #
  # @method getAction
  # @memberof stout-auth/provider/PermissionsProvider#
  ###
  getAction: (action) ->
    fn = @__actions[action]
    if not fn
      throw new err.Err "The action \"#{action}\" is not defined."
    fn


  ###*
  # Tests if the specified action is allowed given the passed set of
  # permissions.
  #
  # @method testAction
  # @memberof stout-auth/provider/PermissionsProvider
  ###
  testAction: (action, permissions) ->
    @getAction(action)(permissions)
