###*
# @overview Defines the default server-side authentication services class which
# accepts and processes authentication requests from a client using the
# `stout-auth/service/ClientAuthService` service.
#
# @module stout-auth/service/ServerAuthService
###
AuthService   = require './AuthService'
crypto        = require 'crypto'
err           = require '../err'
Password      = require '../model/Password'
PasswordReset = require '../model/Password'
Promise       = require 'stout-core/promise/Promise'
Scrypt        = require '../password/Scrypt'



###*
# The ServerAuthService class provides server-side authentication services for
# web applications. It's designed to seamlessly handle requests originating
# from clients running the corresponding client-side service,
# `stout-auth/service/ClientAuthService`.
#
# @exports stout-auth/service/ServerAuthService
# @extends stout-auth/service/AuthProvider
# @constructor
###
module.exports = AuthService.extend 'ServerAuthService',

  constructor: ->
    AuthService.prototype.constructor.apply @, arguments
    @hasher = new Scrypt()
    @handlers =
      'signin': @signin
      'signout': @signout
      'generate-verification-token': @generateVerificationToken
      'check-verification-token': @checkVerificationToken
      'register': @register
      'request-password-reset': @requestPasswordReset
      'reset-password': @resetPassword
      'change-password': @changePassword
    return


  ###*
  # Checks if the `password` matches the user's password.
  #
  # @param {stout-auth/model/User} user - The user model.
  #
  # @param {string} password - The password to check.
  #
  # @returns {stout-core/promise/Promise} Promise resolved to `true` or `false`,
  # indicating if the password matches, or rejected if the user does not have
  # an associated password
  #
  # @method _checkPassword
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _checkPassword: (user, rawPassword) ->
    user.password.get().then (password) =>
      if not password
        throw new err.AuthErr 'User does not have a password.'
      else
        @_getHasher(password).verify(password.verifier, rawPassword)


  ###*
  # Returns the appropriate hasher for the passed password object. Currently,
  # this method simply returns the only hasher, but it could be extended in the
  # future to return one of multiple hashers if the database contains passwords
  # hashed over time with separate passwords.
  #
  # @method _getHasher
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _getHasher: (password) -> @hasher


  ###*
  # Returns the reduced precision clock for generating and checking
  # verification tokens.
  #
  # @returns {number} Reduced precision clock.
  #
  # @method _getVerificationClock
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _getVerificationClock: ->
    config = @providers.get('config')
    period = config.get('auth.verification.period') * 60 * 1000 / 2
    parseInt(Date.now() / period)


  ###*
  # Makes a new password reset model for the passed user.
  #
  # @param {stout-auth/model/User} user - The user to generate the password
  # reset for.
  #
  # @returns {stout-auth/model/PasswordReset} The password reset model.
  #
  # @method _makePasswordReset
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _makePasswordReset: (user) ->
    config = @providers.get('config')
    secret = config.get('auth.secret')
    expires = config.get('auth.reset.expires')
    hmac = crypto.createHmac('sha256', secret)
    hmac.update user.key
    new PasswordReset
      token: hmac.digest('hex')
      expires: Date.now() + expires * 60 * 1000


  ###*
  # Generates a verification token for the passed user and session.
  #
  # @param {stout-auth/model/User} user - The user model.
  #
  # @param {} session - The user's current session.
  #
  # @returns {string} The verification token.
  #
  # @method _makeVerificationToken
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _makeVerificationToken: (user, session, clock) ->
    config = @providers.get('config')
    secret = config.get('auth.secret')
    hmac = crypto.createHmac('sha256', secret)
    hmac.update user.key + clock + session.key
    hmac.digest('hex')


  ###*
  # Generates a new password verifier and returns a Password model.
  #
  # @param {string} password - The password to hash.
  #
  # @returns {stout-core/promise/Promise} Promise resolved with Password model.
  #
  # @method _makePasswordVerifier
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _makePasswordVerifier: (password) ->
    @hasher.generate(password).then (verifier) ->
      pass = new Password
        algorithm: 'scrypt'
        version: 1
        verifier: verifier
      pass.save().then -> pass



  ###*
  # Sends the password reset request email to the user.
  #
  # @param {stout-auth/model/User} user - The user to send the password
  # reset request email to.
  #
  # @param {stout-auth/model/passwordReset} passwordReset - The password reset
  # model.
  #
  # @method _sendPasswordResetEmail
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _sendPasswordResetEmail: (user, passwordReset) ->
    @services.request 'email:send',
      template: 'emails.password-reset'
      date: {user, passwordReset}
      to: user.email


  ###*
  # Creates a password verifier and attaches it to the passed user model.
  #
  # @param {stout-auth/model/User} user - The user model.
  #
  # @param {string} password - The password.
  #
  # @method _setUserPassword
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _setUserPassword: (user, password) ->
    @_makePasswordVerifier(password).then (verifier) ->
      user.password.set(verifier).then ->
        user.save()
      .catch (error) ->
        verifier.delete()
        throw error


  ###*
  # Starts a session for the passed user.
  #
  # @method _startSession
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _startSession: (user) ->
    #@services.request('session:start', user)
    Promise.fulfilled()


  ###*
  # Ends a session for the passed user.
  #
  # @method _endSession
  # @memberof stout-auth/server/AuthProvider#
  # @protected
  ###
  _endSession: (user) ->
    #@services.request('session:end', user)
    Promise.fulfilled()


  ###*
  # Changes a user's password.
  #
  # @param {function} fulfill - Function to call after a successful password
  # change.
  #
  # @param {function} reject - Function to call with a failed password change.
  #
  # @param {stout-auth/model/User} user - The user object.
  #
  # @param {string} currentPassword - The user's current password.
  #
  # @param {string} password - The new password.
  #
  # @method changePassword
  # @memberof stout-auth/server/AuthProvider#
  ###
  changePassword: (fulfill, reject, user, currentPassword, newPassword) ->
    username = user.getUsername()
    @log.debug 'Changing password for user %s.', username, user
    @fire 'change-password:request', user
    @_checkPassword(user, currentPassword).then (matches) ->
      if matches
        @_setUserPassword(user, newPassword).then ->
          @log.info 'Successful changed password for user %s.', username, user
          @fire 'change-password:success', user
          fulfill()
      else
        @log.notice 'Failed to change password for user %s. Current password
          does not match.', username, user
        error = new err.PasswordErr 'Invalid password.'
        @fire 'change-password:error', {user, error}
        reject error
    .catch (error) ->
      @log.warning 'Error occurred attempting to change password for user %s.',
        username, {user, error}
      @fire 'change-password:error', {user, error}
      reject error


  ###*
  # Checks a verification token. If the token is valid, the request is
  # fulfilled with an updated token.
  #
  # @param {function} fulfill - Function to call with a successful verification.
  #
  # @param {function} reject - Function to call with a failed verification.
  #
  # @param {stout-auth/model/User} user - The user object.
  #
  # @param {string} token - The token to check.
  #
  # @method checkVerificationToken
  # @memberof stout-auth/server/AuthProvider#
  ###
  checkVerificationToken: (fulfill, reject, user, token) ->
    user.session.get().then (session) ->
      if not session
        @log.warning 'Cannot check token for user %s.
          No valid session exists.', username, user
        error = new err.VerificationErr 'No valid session exists.'
        @fire 'verify:error', {user, error}
      else
        clock = @_getVerificationClock()
        vToken = @_makeVerificationToken(user, session, clock)
        vTokenPrevious = @_makeVerificationToken(user, session, clock - 1)
        if vToken is token or vTokenPrevious is token
          @log.info 'Validated verification token for user %s.', user
          @fire 'verify:success', {user, token}
          fulfill vToken
        else
          @log.info 'Invalid verification token for user %s.', user
          @fire 'verify:success', {user, token}
          fulfill()
    .catch (error) ->
      @log.warning 'Error occurred attempting to check verification
        token for %s.', username, {user, error}
      @fire 'verify:error', {user, error}
      reject error


  ###*
  # Verifies the passed credentials and returns a token and timestamp for later
  # use to ensure the user did in-fact verify their credentials within a given
  # amount of time.
  #
  # Here's how this works:
  # 1. Alice performs some operation (client side) which requires that she
  #    verify her password (i.e. change critical settings, delete account,
  #    etc.).
  # 2. Alice is presented with a "verify password" modal or similar UI element
  #    where she enters her password.
  # 3. The client-side system sends the password to the server.
  # 4. The server checks if the password matches Alice's stored password.
  # 5. If the passwords match, a token is generated as follows:
  #    a. Let `id` be Alice's user ID (UUID), `t` be the current timestamp,
  #       and `secret` be a secret key stored on the server.
  #    b. Let `k = HMAC(secret, id + t)`
  #    c. Generate a secure token as `token = k + t`,
  #       where `+` is the concatenation operator.
  # 6. Send the token to Alice.
  # 7. For each "dangerous" operation, Alice includes `token` with the request.
  # 8. The server can verify the `token` by splitting `k` and `t` and
  # validating `k` as in step 5.b above, and verify that time `T` hasn't
  # elapsed since Alice verified her password.
  #
  # Alice can then include the token with future dangerous requests to validate
  # that she has verified her password, and that it was verified at a specific
  # time. The time can be used to ensure that when performing dangerous
  # operations, the user has verified their identity within the past X minutes.
  #
  # @param {object} [request] - Credentials verification request.
  #
  # @param {function} fulfill - Function to call with a successful verify.
  #
  # @param {function} reject - Function to call with a failed verify.
  #
  # @method generateVerificationToken
  # @memberof stout-auth/server/AuthProvider#
  ###
  generateVerificationToken: (fulfill, reject, user, password) ->
    @_checkPassword(user, password).then (matches) ->
      if matches
        user.session.get().then (session) ->
          if not session
            @log.warning 'Cannot generate token for user %s.
              No valid session exists.', username, user
            error = new err.VerificationErr 'No valid session exists.'
            @fire 'verify:token:error', {user, error}
          else
            clock = _getVerificationClock()
            token = @_makeVerificationToken(user, session, clock)
            @log.info 'Generated verification token for %s.', user
            @fire 'verify:token:success', {user, token}
            fulfill token
      else
        @log.notice 'Failed verification token generation attempt.
        Incorrect password for user %s.', username, user
        error = new err.PasswordErr 'Invalid password.'
        @fire 'verify:token:error', {user, error}
        reject error
    .catch (error) ->
      @log.warning 'Error occurred attempting to generate verification
        token for %s.', username, {user, error}
      @fire 'verify:token:error', {user, error}
      reject error


  ###*
  # Registers a new user by hashing the password, then saving the user and
  # linking the password verifier.
  #
  # @param {object} [request] - Registration request.
  #
  # @param {function} fulfill - Function to call with a successful registration.
  #
  # @param {function} reject - Function to call with a failed registration.
  #
  # @method register
  # @memberof stout-auth/server/AuthProvider#
  ###
  register: (fulfill, reject, user, password) ->
    username = user.getUsername()
    @log.debug 'Registering user %s.', username, user
    @fire 'register:request'
    @_makePasswordVerifier(password).then (verifier) =>
      user.password.link(verifier).then =>
        user.save().then =>
          @log.info 'Successfully registered user %s.', username, user
          @fire 'register:success', user
          fulfill()
      .catch (error) =>
        @log.warning 'Error occurred registering %s.', username, {user, error}
        @fire 'register:error', {user, error}
        verifier.delete()
        user.delete()
        reject(error)
    .catch (error) =>
      @log.warning 'Error generating password verifier for %s.',
        username, {user, error}
      @fire 'register:error', {user, error}
      reject(error)


  ###*
  # Deletes any pending password resets for the passed user, generates a new
  # password reset request, and emails the user the password reset email.
  #
  # @param {function} fulfill - Function to call with a successful password
  # reset request.
  #
  # @param {function} reject - Function to call with a failed password reset
  # request.
  #
  # @param {stout-auth/model/User} user - The user object.
  #
  # @method requestPasswordReset
  # @memberof stout-auth/server/AuthProvider#
  ###
  requestPasswordReset: (fulfill, reject, user) ->
    username = user.getUsername()
    user.passwordReset.get().then (previousReset) ->
      if previousReset
        previousReset.delete().catch (error) ->
          @log.error 'Error deleting previous password reset for user %s.',
            username, {user, error}
      passwordReset = @_makePasswordReset(user)
      user.passwordReset.set(passwordReset).then ->
        @_sendPasswordResetEmail(user, passwordReset).then ->
          @log.info 'Sent password reset email to user %s.', username, user
          fulfill()
        .catch (error) ->
          @log.warning 'Error sending password reset email to user %s.',
            username, {user, error}
          # @fire 'password-reset:error', {user, error}
          reject error
    .catch (error) ->
      @log.warning 'Error generating password reset for user %s.',
        username, {user, error}
      # @fire 'password-reset:error', {user, error}
      reject error


  ###*
  # Resets the user's password.
  #
  # @param {function} fulfill - Function to call after a successful password
  # reset.
  #
  # @param {function} reject - Function to call with a failed password reset.
  #
  # @param {stout-auth/model/User} user - The user object.
  #
  # @param {string} token - Password reset request token.
  #
  # @param {string} password - The new password.
  #
  # @method resetPassword
  # @memberof stout-auth/server/AuthProvider#
  ###
  resetPassword: (fulfill, reject, user, token, password) ->
    username = user.getUsername()
    @log.debug 'Resetting password for user %s.', username, user
    # @fire 'register:request'
    user.passwordReset.get().then (passwordReset) ->
      if not passwordReset or
      passwordReset.expired() or
      passwordReset.token not token
        @log.notice 'Invalid password token or token has expired for user %s',
          username, user
        error = new err.PasswordResetErr 'Invalid or expired password
        reset token.'
        reject error
      else
        @_setUserPassword(user, password).then ->
          @log.info 'Successfully reset password for user %s.', username, user
          @fire 'change-password:success', user
          fulfill()
    .catch (error) ->
      @log.warning 'Error reseting password for user %s.',
        username, {user, error}
      # @fire 'register:error', {user, error}
      reject error


  ###*
  # Signs-in a user, verifying the passed credentials.
  #
  # @param {object} [request] - Sign-in request.
  #
  # @param {function} fulfill - Function to call with a successful sign-in
  # request.
  #
  # @param {function} reject - Function to call with a failed sign-in
  # request.
  #
  # @method signin
  # @memberof stout-auth/server/AuthProvider#
  ###
  signin: (fulfill, reject, user, password) ->
    username = user.getUsername()
    @log.debug 'Signing-in user %s.', username, user
    @fire 'signin:request', user
    @_checkPassword(user, password).then (matches) =>
      if matches
        @log.info 'Successful sign in for %s.', username, user
        @fire 'signin:success', user
        @_startSession(user).then -> fulfill()
      else
        @log.notice 'Failed sign in attempt. Incorrect password for user %s.',
          username, user
        error = new err.PasswordErr 'Invalid password.'
        @fire 'signin:error', {user, error}
        reject error
    .catch (error) =>
      @log.warning 'Error occurred attempting to sign in %s.',
        username, {user, error}
      @fire 'signout:error', {user, error}
      reject(error)


  ###*
  # Signs-out a user.
  #
  # @param {object} [request] - Sign-out request.
  #
  # @param {function} fulfill - Function to call with a successful sign-out.
  #
  # @param {function} reject - Function to call with a failed sign-out.
  #
  # @method signout
  # @memberof stout-auth/server/AuthProvider#
  ###
  signout: (fulfill, reject, user) ->
    username = user.getUsername()
    @log.debug 'Signing-out user %s.', username, user
    @fire 'signout:request'
    @_endSession(user).then ->
      @log.info 'Successful sign out for %s.', username, user
      @fire 'signout:success'
      fulfill()
    .catch (error) ->
      @log.warning 'Error occurred attempting to sign out %s.',
        username, {user, error}
      @fire 'signout:error', {user, error}
      reject(error)
