assert              = require('chai').assert
Datastore           = require 'stout-memory-datastore/Datastore'
JoinModel           = require 'stout-orm/model/JoinModel'
ModelProvider       = require 'stout-orm/provider/ModelProvider'
Password            = require '../../../model/Password'
PermissionsProvider = require '../../../provider/PermissionsProvider'
Role                = require '../../../model/Role'
RoleProvider        = require '../../../provider/RoleProvider'
sinon               = require 'sinon'
User                = require '../../../model/User'



describe 'model/User', ->
  user = permissionsProvider = null

  before ->
    Role.__linked = false
    User.__linked = false
    Password.__linked = false

    # Define default join provider for many-to-many relationships.
    JoinModel.provider = new ModelProvider
      datastore: new Datastore,
      model: JoinModel

    # Wire-up the role provider.
    roleProvider = new RoleProvider
      datastore: new Datastore
      model: Role
    Role.provider = roleProvider
    Role.hasMany = [User]

    Password.provider = null
    Password.datastore = new Datastore
    Password.belongsTo = [User]
    Password.$boot()

    User.provider = null
    User.datastore = new Datastore
    User.hasOne = [Password]
    User.hasMany = [Role] # TODO: Add stout-session for testing.

    # Mock permissions provider.
    permissionsProvider = new PermissionsProvider

    Role.$boot()
    User.$boot {providers: {get: -> permissionsProvider}}

  beforeEach ->
    user = new User
      email: 'test@example.com'


  for prop in ['email', 'permissions']
    ((prop) ->
      it "has a ##{prop} property", ->
        assert.property user, prop
    )(prop)


  describe '#can()', ->

    it 'resolves to `true` if user has passed permission', ->
      user.permissions.push 'add-users'
      user.granted('add-users').then (canAddUsers) ->
        assert.isTrue canAddUsers

    it 'tests action with permissions provider', ->
      permissionsProvider.defineAction 'add admin users', ['a', 'b', 'c']
      user.grant 'a', ['b', 'c'], 'd'
      user.can('add admin users').then (canAddAdminUsers) ->
        assert.isTrue canAddAdminUsers


  describe '#getUsername()', ->

    it 'returns the email property by default', ->
      assert.equal user.getUsername(), 'test@example.com'



  describe '#getAllPermissions()', ->

    it 'returns empty array when no roles or permissions are defined', ->
      user.getAllPermissions().then (permissions) ->
        assert.deepEqual permissions, []

    it 'returns locally defined permissions', ->
      user.permissions.push 'test-perm'
      user.permissions.push 'another-perm'
      user.getAllPermissions().then (permissions) ->
        assert.sameDeepMembers permissions, ['test-perm', 'another-perm']

    it 'returns permissions defined in role', ->
      user.roles.createAndSave({name: 'test-role', permissions: ['a', 'b']})
      .then ->
        user.getAllPermissions().then (permissions) ->
          assert.sameDeepMembers permissions, ['a', 'b']

    it 'returns union of role and specific permissions', ->
      user.permissions.push 'c'
      user.roles.createAndSave({name: 'test-role', permissions: ['a', 'b']})
      .then ->
        user.getAllPermissions().then (permissions) ->
          assert.sameDeepMembers permissions, ['a', 'b', 'c']

    it 'multiple users may share the same role', ->
      user.roles.createAndSave({name: 'test-role', permissions: ['a', 'b']})
      .then (role) ->
        user2 = new User
          email: 'test2@example.com'
          permissions: ['x', 'y']
        user2.roles.link(role).then ->
          user.getAllPermissions().then (permissions) ->
            assert.sameDeepMembers permissions, ['a', 'b']
            user2.getAllPermissions().then (permissions) ->
              assert.sameDeepMembers permissions, ['a', 'b', 'x', 'y']


  describe '#grant()', ->

    it 'grants the user the passed permissions', ->
      user.grant 'a'
      user.granted('a').then (can) ->
        assert.isTrue can

    it 'grants var-arg permissions to user', ->
      user.grant 'a', 'b', ['c', 'x'], 'y'
      permissionsProvider.defineAction 'test action', ['a', 'b', 'c']
      user.can('test action').then (can) ->
        assert.isTrue can


  describe '#revoke()', ->

    it 'revokes permissions from the user', ->
      user.grant 'a'
      user.granted('a').then (can) ->
        assert.isTrue can
        user.revoke 'a'
        user.granted('a').then (can) ->
          assert.isFalse can
