assert            = require('chai').assert
ConfigProvider    = require 'stout-core/config/ConfigProvider'
ConsoleDrain      = require 'stout-core/log/drain/Console'
Datastore         = require 'stout-memory-datastore/Datastore'
Password          = require '../../../model/Password'
PasswordProvider  = require '../../../provider/PasswordProvider'
PermissionsProvider = require '../../../provider/PermissionsProvider'
ProviderSet       = require 'stout-core/provider/ProviderSet'
ServerAuthService = require '../../../service/ServerAuthService'
Session           = require 'stout-session/model/Session'
SessionProvider   = require 'stout-session/provider/SessionProvider'
sinon             = require 'sinon'
User              = require '../../../model/User'
UserProvider      = require '../../../provider/UserProvider'



describe 'service/ServerAuthService', ->
  auth = s1 = null

  beforeEach ->
    Password.__linked = false
    User.__linked = false
    Session.__linked = false

    config = new ConfigProvider
    config.set 'auth.secret', 'secret'
    config.set 'auth.reset.expire', 1
    config.set 'auth.verification.period', 5

    Password.provider = new PasswordProvider
      datastore: new Datastore
      model: Password
    Password.belongsTo = [User]

    User.provider = new UserProvider
      datastore: new Datastore
      model: User
    User.hasOne = [Password]
    User.hasMany = []

    Session.provider = new SessionProvider
      datastore: new Datastore
      model: Session
    Session.hasOne = []

    Password.$boot()
    Session.$boot()

    # Mock permissions provider.
    permissionsProvider = new PermissionsProvider
    User.$boot {providers: {get: -> permissionsProvider}}

    auth = new ServerAuthService
    auth.providers = new ProviderSet
    auth.providers.add config

    #auth.log.drains.add new ConsoleDrain



  for handler in [
    'signin'
    'signout'
    'generate-verification-token'
    'check-verification-token'
    'register'
    'request-password-reset'
    'reset-password'
    'change-password'
  ]
    ((handler) ->
      it "has a \"#{handler}\" handler", ->
        assert.property auth.handlers, handler
    )(handler)



  describe '#_makePasswordVerifier', ->

    it 'returns a Password object', ->
      auth._makePasswordVerifier('test-password').then (password) ->
        assert.instanceOf password, Password

    it 'generates a password verifier using the current hashing function', ->
      auth._makePasswordVerifier('test-password').then (password) ->
        assert.isAtLeast Buffer(password.verifier, 'base64').length, 82

    it 'sets the Password object\'s algorithm and version correctly', ->
      auth._makePasswordVerifier('test-password').then (password) ->
        assert.equal password.algorithm, 'scrypt'
        assert.equal password.version, 1



  describe '#_makeVerificationToken', ->

    it 'returns a sha256 token', ->
      token = auth._makeVerificationToken {key: 'a'}, {key: 'b'}, 60000
      assert.isString token
      assert.lengthOf token, 64



  describe '#register()', ->

    it 'registers a new user', ->
      user = new User email: 'test@example.com'
      new Promise (f, r) ->
        auth.register f, r, user, 'fake password'
      .then ->
        User.provider.find({email: 'test@example.com'})
        .then (results) ->
          assert.equal results[0].key, user.key



  describe '#signin()', ->
    userEmail = 'test@example.com'
    userPass = 'fake password'

    beforeEach ->
      user = new User email: userEmail
      new Promise (f, r) ->
        auth.register f, r, user, userPass

    it 'signs in a user', ->
      User.provider.find({email: userEmail}).then (results) ->
        user = results[0]
        new Promise (f, r) ->
          auth.signin f, r, user, userPass

    it 'rejects for incorrect password', (done) ->
      User.provider.find({email: userEmail}).then (results) ->
        user = results[0]
        new Promise (f, r) ->
          auth.signin f, r, user, 'incorrect password'
        .then -> done new Error 'Expected failure from incorrect password.'
        .catch -> done()
      return

    it 'rejects if the user does not have a password', (done) ->
      User.provider.find({email: userEmail}).then (results) ->
        user = results[0]
        user.password.get().then (pass) ->
          pass.delete().then ->
            new Promise (f, r) ->
              auth.signin f, r, user, userPass
            .then -> done new Error 'Expected failure from missing password.'
            .catch -> done()
      return



  describe '#signout()', ->
    userEmail = 'test@example.com'
    userPass = 'fake password'

    beforeEach ->
      user = new User email: userEmail
      new Promise (f, r) ->
        auth.register f, r, user, userPass
